<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub inbox requests
 */
class Inbox implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main api main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch inbox for @unused-param
     * @return string|false response
     */
    public function get($_user)
    {
        return false;
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to inbox
     * @return string|false response
     */
    public function post($_user)
    {
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            throw new \Federator\Exceptions\Unauthorized("Inbox::post Signature check failed: " . $e->getMessage());
        }

        $activity = is_string($_rawInput) ? json_decode($_rawInput, true) : null;

        if (!is_array($activity)) {
            throw new \Federator\Exceptions\ServerError("Inbox::post Input wasn't of type array");
        }

        $inboxActivity = \Federator\Data\ActivityPub\Factory::newActivityFromJson($activity);

        if ($inboxActivity === false) {
            throw new \Federator\Exceptions\ServerError("Inbox::post couldn't create inboxActivity");
        }

        $rootDir = PROJECT_ROOT . '/';

        // Shared inbox
        if (!isset($_user)) {
            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/inbox.log',
                date('Y-m-d H:i:s') . ": ==== POST Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }

        $sendTo = $inboxActivity->getCC();
        if ($inboxActivity->getType() === 'Undo') { // for undo the object holds the proper cc
            $object = $inboxActivity->getObject();
            if ($object !== null && is_object($object)) {
                $sendTo = $object->getCC();
            }
        }

        $users = [];
        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        foreach ($sendTo as $receiver) {
            if ($receiver === '' || !is_string($receiver)) {
                continue;
            }

            if (str_ends_with($receiver, '/followers')) {
                $actor = $inboxActivity->getAActor();
                if ($actor === null || !is_string($actor)) {
                    error_log("Inbox::post no actor found");
                    continue;
                }

                // Extract username from the actor URL
                $username = basename((string)(parse_url($actor, PHP_URL_PATH) ?? ''));
                $domain = parse_url($actor, PHP_URL_HOST);
                if ($username === null || $domain === null) {
                    error_log("Inbox::post no username or domain found");
                    continue;
                }
                $followers = \Federator\DIO\Followers::getFollowersByFedUser($dbh, $connector, $cache, $username . '@' . $domain);

                if (is_array($followers)) {
                    $users = array_merge($users, array_column($followers, 'id'));
                }
            }
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }
        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }

            $token = \Resque::enqueue('inbox', 'Federator\\Jobs\\InboxJob', [
                'user' => $user,
                'activity' => $inboxActivity->toObject(),
            ]);
            error_log("Inbox::post enqueued job for user: $user with token: $token");
        }
        return "success";
    }

    /**
     * handle post call for specific user
     *
     * @param \mysqli $dbh database handle
     * @param \Federator\Connector\Connector $connector connector to use
     * @param \Federator\Cache\Cache|null $cache optional caching service
     * @param string $_user user to add data to inbox
     * @param \Federator\Data\ActivityPub\Common\Activity $inboxActivity the activity that we received
     * @return boolean response
     */
    public static function postForUser($dbh, $connector, $cache, $_user, $inboxActivity)
    {
        if (isset($_user)) {
            // get user
            $user = \Federator\DIO\User::getUserByName(
                $dbh,
                $_user,
                $connector,
                $cache
            );
            if ($user === null || $user->id === null) {
                throw new \Federator\Exceptions\ServerError("Inbox::postForUser couldn't find user: $_user");
            }
        }

        $rootDir = PROJECT_ROOT . '/';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/inbox_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        return true;
    }
}
