<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\V1;

/**
 * Called from our application to inform us about new content (f.e. new posts on contentnation.net)
 */
class NewContent implements \Federator\Api\APIInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * response from sub-calls
     *
     * @var string $response
     */
    private $response;

    /**
     * constructor
     *
     * @param \Federator\Main $main main instance
     * @return void
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * run given url path
     *
     * @param array<string> $paths path array split by /
     * @param \Federator\Data\User|false $user user who is calling us @unused-param
     * @return bool true on success
     */
    public function exec($paths, $user)
    {
        $method = $_SERVER["REQUEST_METHOD"];
        $_username = $paths[2];
        if ($method === 'GET') { // unsupported
            throw new \Federator\Exceptions\InvalidArgument("GET not supported");
        }
        switch (sizeof($paths)) {
            case 3:
                $ret = $this->post($_username);
                break;
        }

        if (isset($ret) && $ret !== false) {
            $this->response = $ret;
            return true;
        }

        $this->main->setResponseCode(404);
        return false;
    }


    /**
     * handle post call
     *
     * @param string|null $_user optional user that triggered the post
     * @return string|false response
     */
    public function post($_user)
    {
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            error_log("NewContent::post Signature check failed: " . $e->getMessage());
            http_response_code(401);
            return false;
        }

        $input = is_string($_rawInput) ? json_decode($_rawInput, true) : null;

        $config = $this->main->getConfig();
        $domain = $config['generic']['externaldomain'];
        if (!is_array($input)) {
            error_log("NewContent::post Input wasn't of type array");
            return false;
        }

        if (isset($allHeaders['X-Sender'])) {
            $newActivity = $this->main->getConnector()->jsonToActivity($input);
        } else {
            error_log("NewContent::post No X-Sender header found");
            return false;
        }

        if ($newActivity === false) {
            error_log("NewContent::post couldn't create newActivity");
            return false;
        }

        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        if (!isset($_user)) {
            $user = $newActivity->getAActor(); // url of the sender https://contentnation.net/username
            $user = str_replace(
                $domain,
                '',
                $user
            ); // retrieve only the last part of the url
        } else {
            $user = $dbh->real_escape_string($_user);
        }

        $users = [];

        $followers = $this->fetchAllFollowers($dbh, $connector, $cache, $user);
        if (!empty($followers)) {
            $users = array_merge($users, $followers);
        }

        if (empty($users)) { // todo remove after proper implementation, debugging for now
            $rootDir = PROJECT_ROOT . '/';
            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/newContent.log',
                date('Y-m-d H:i:s') . ": ==== POST NewContent Activity ====\n" . json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }

        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }

            $this->postForUser($dbh, $connector, $cache, $user, $newActivity);
        }

        return json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call for specific user
     *
     * @param \mysqli $dbh @unused-param
     *          database handle
     * @param \Federator\Connector\Connector $connector
     *          connector to fetch use with
     * @param \Federator\Cache\Cache|null $cache
     *          optional caching service
     * @param string $_user user that triggered the post
     * @param \Federator\Data\ActivityPub\Common\Activity $newActivity the activity that we received
     * @return boolean response
     */
    private static function postForUser($dbh, $connector, $cache, $_user, $newActivity)
    {
        if (!isset($_user)) {
            error_log("NewContent::postForUser no user given");
            return false;
        }

        // get user
        $user = \Federator\DIO\User::getUserByName(
            $dbh,
            $_user,
            $connector,
            $cache
        );
        if ($user->id === null) {
            error_log("NewContent::postForUser couldn't find user: $_user");
            return false;
        }

        $rootDir = PROJECT_ROOT . '/';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/newcontent_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " NewContent Activity ====\n" . json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        return true;
    }

    /**
     * fetch all followers from url and return the ones that belong to our server
     *
     * @param \mysqli $dbh
     *          database handle
     * @param \Federator\Connector\Connector $connector
     *          connector to fetch use with
     * @param \Federator\Cache\Cache|null $cache
     *          optional caching service
     * @param string $userId The id of the user
     * @return string[] the names of the followers that are hosted on our server
     */
    private static function fetchAllFollowers($dbh, $connector, $cache, string $userId): array
    {
        if (empty($userId)) {
            return [];
        }

        $users = [];

        $apFollowers = \Federator\DIO\Followers::getFollowersByUser(
            $dbh,
            $userId,
            $connector,
            cache: $cache,
        );

        foreach ($apFollowers as $follower) {
            $users[] = $follower->id;
        }

        return $users;
    }

    /**
     * get internal represenation as json string
     * @return string json string or html
     */
    public function toJson()
    {
        return $this->response;
    }
}